import socket
import uuid

from typing import Callable, List, Tuple

from .validation import XmppValidation
from .permission import XmppPermission
from .message import XmppMessage
from .features import XmppFeature
from .extensions import XmppExtension
from .ci import XmppClientInterface

from osmxml import *

import logging


logger = logging.getLogger(__name__)


class XmppClient:
    """
    XMPP client implementation.
    """

    def __init__(self, host:str, port:int=5222):
        """
        Initializes the XMPP client.

        Args:
            host (str): The host of the XMPP server.
            port (int): The port of the XMPP server.
        """
        self.host = host
        self.port = port

        self.__connected = False

        self.__hooks = {
            "send_message": [],
            "on_message": [],
            "on_presence": [],
            "on_iq": [],
        }

        self.__handlers = {
            "connected": [],
            "disconnected": [],
            "ready": [],
            "message": [],
            "presence": [],
            "iq": [],
        }

        self.__features = {}
        self.__features_queue = []

        self.__extensions = {}
        self.__extensions_queue = []


    @property    
    def extensions(self):
        extensions_functions = {}
        for extension_id, extension_ci in self.__extensions.items():
            extensions_functions[extension_id] = extension_ci.variables
            
        return extensions_functions


    @property
    def connected(self):
        return self.__connecte
    

    def _trigger_handlers(self, event:str, *args, **kwargs):
        logger.debug(f"Triggering '{event}' handlers...")
        for handler in self.__handlers[event]:
            handler(*args, **kwargs)
    
    def _trigger_hooks(self, event:str, value, *args, **kwargs):
        logger.debug(f"Triggering '{event}' hooks...")
        for hook in self.__hooks[event]:
            value = hook(value, *args, **kwargs)
            if not value:
                return None
        return value

    
    def send_message(self, *args, **kwargs):
        """
        Sends a message to the given JID.

        Args:
            jid (str): The JID to send the message to.
            message (str): The message to send.
            type (str): The message type. (Default: "chat")
        
        Example:
            >>> client.send_message("john@jabber.org", "Hello, John!")
        """

        jid = args[0] if len(args) > 0 else kwargs.get("jid")
        content = args[1] if len(args) > 1 else kwargs.get("message")
        msg_type = kwargs.get("type", "chat")

        XmppValidation.validate_jid(jid)

        message = XmppMessage()
        message.xml.add_attribute(XmlAttribute("to", jid))
        message.xml.add_attribute(XmlAttribute("type", msg_type))
        message.xml.add_attribute(XmlAttribute("id", str(uuid.uuid4())))

        message.xml.add_child(XmlElement("body"))
        message.body.xml.add_child(XmlTextElement(content))

        for hook in self.__hooks["send_message"]:
            message = hook(message, *args, **kwargs)

        if not message:
            return
        self._send_xml(message.xml)
    
    def reply_to_message(self, *args, **kwargs):
        """
        Replies to specific message.

        Args:
            message_id (str): The message ID to reply to.
            jid (str): The JID to send the message to.
            message (str): The message to send.
            type (str): The message type. (Default: "chat")
            message_author (str): The replied message author. If none, equals to jid. (Default: None)
        
        Example:
            >>> client.reply_to_message("12345678", "john@jabber.org", "Thanks, John!")
        """

        message_id = args[0] if len(args) > 0 else kwargs.get("message_id")
        jid = args[1] if len(args) > 1 else kwargs.get("jid")
        content = args[2] if len(args) > 2 else kwargs.get("message")
        msg_type = kwargs.get("type", "chat")
        message_author = kwargs.get("message_author", None)

        XmppValidation.validate_jid(jid)

        if message_author:
            XmppValidation.validate_jid(message_author)

        message = XmppMessage()
        message.xml.add_attribute(XmlAttribute("to", jid))
        message.xml.add_attribute(XmlAttribute("type", msg_type))
        message.xml.add_attribute(XmlAttribute("id", str(uuid.uuid4())))

        message.xml.add_child(XmlElement("body"))
        message.body.xml.add_child(XmlTextElement(content))

        message._xml.add_child(XmlElement("reply"))
        message.reply.xml.add_attribute(XmlAttribute("xmlns", "urn:xmpp:reply:0"))
        message.reply.xml.add_attribute(XmlAttribute("id", message_id))
        if message_author:
            message.reply.xml.add_attribute(XmlAttribute("to", message_author))
        else:
            message.reply.xml.add_attribute(XmlAttribute("to", jid))

        for hook in self.__hooks["send_message"]:
            message = hook(message, *args, **kwargs)

        if not message:
            return
        self._send_xml(message.xml)

    def edit_message(self, *args, **kwargs):
        """
        Editing specific message.

        Args:
            message_id (str): The message ID to edit.
            jid (str): The JID to send the message to.
            message (str): The message to send.
            type (str): The message type. (Default: "chat")
        
        Example:
            >>> client.edit_message("12345679", "john@jabber.org", "Thank you very much, John!")
        """

        message_id = args[0] if len(args) > 0 else kwargs.get("message_id")
        jid = args[1] if len(args) > 1 else kwargs.get("jid")
        content = args[2] if len(args) > 2 else kwargs.get("message")
        msg_type = kwargs.get("type", "chat")

        XmppValidation.validate_jid(jid)

        message = XmppMessage()
        message.xml.add_attribute(XmlAttribute("to", jid))
        message.xml.add_attribute(XmlAttribute("type", msg_type))
        message.xml.add_attribute(XmlAttribute("id", str(uuid.uuid4())))

        message.xml.add_child(XmlElement("body"))
        message.body.xml.add_child(XmlTextElement(content))

        message.xml.add_child(XmlElement("replace"))
        message.replace.xml.add_attribute(XmlAttribute("xmlns", "urn:xmpp:message-correct:0"))
        message.replace.xml.add_attribute(XmlAttribute("id", message_id))

        for hook in self.__hooks["send_message"]:
            message = hook(message, *args, **kwargs)

        if not message:
            return
        self._send_xml(message.xml)


    def on_connect(self, handler:Callable) -> Callable:
        """
        Registers a handler for the connected event.
        The handler will be called when the client is connected to the XMPP server, but not ready yet.

        Args:
            handler (Callable): The handler to register.

        Returns:
            Callable: The handler (not changed).
        """
        self.__handlers["connected"].append(handler)
        return handler
    
    def on_disconnect(self, handler:Callable) -> Callable:
        """
        Registers a handler for the disconnected event.
        The handler will be called when the client is disconnected from the XMPP server.

        Args:
            handler (Callable): The handler to register.

        Returns:
            Callable: The handler (not changed).
        """
        self.__handlers["disconnected"].append(handler)
        return handler
    
    def on_ready(self, handler:Callable) -> Callable:
        """
        Registers a handler for the ready event.
        The handler will be called when the client is ready to send and receive XMPP stanzas.

        Args:
            handler (Callable): The handler to register.

        Returns:
            Callable: The handler (not changed).
        
        Example:
            >>> @client.on_ready
            ... def on_ready():
            ...     print(f"Loggened in as {client.jid}")
        """
        self.__handlers["ready"].append(handler)
        return handler

    def on_message(self, handler:Callable) -> Callable:
        """
        Registers a handler for the message event.
        The handler will be called when the client receives a message stanza.

        Args:
            handler (Callable): The handler to register.

        Returns:
            Callable: The handler (not changed).

        Example:
            >>> @client.on_message
            ... def on_message(message):
            ...     if message.body is None: #Messages body can be empty
            ...         return
            ...
            ...     print(f"Received message from {message.from_jid}: {message.body}")
        """
        self.__handlers["message"].append(handler)
        return handler
    
    def on_presence(self, handler:Callable) -> Callable:
        """
        Registers a handler for the presence event.
        The handler will be called when the client receives a presence stanza.

        Args:
            handler (Callable): The handler to register.

        Returns:
            Callable: The handler (not changed).
        """
        self.__handlers["presence"].append(handler)
        return handler
    
    def on_iq(self, handler:Callable) -> Callable:
        """
        Registers a handler for the iq event.
        The handler will be called when the client receives an iq stanza.

        Args:
            handler (Callable): The handler to register.

        Returns:
            Callable: The handler (not changed).
        """
        self.__handlers["iq"].append(handler)
        return handler


    def hook_on_message(self, hook:Callable) -> Callable:
        """
        Registers a hook for the message event.
        The hook will be called when the client receives a message stanza.

        Args:
            hook (Callable): The hook to register.

        Returns:
            Callable: The hook (not changed).
        """
        self.__hooks["on_message"].append(hook)
        return hook
    
    def hook_on_presence(self, hook:Callable) -> Callable:
        """
        Registers a hook for the presence event.
        The hook will be called when the client receives a presence stanza.

        Args:
            hook (Callable): The hook to register.

        Returns:
            Callable: The hook (not changed).
        """
        self.__hooks["on_presence"].append(hook)
        return hook
    
    def hook_on_iq(self, hook:Callable) -> Callable:
        """
        Registers a hook for the iq event.
        The hook will be called when the client receives an iq stanza.

        Args:
            hook (Callable): The hook to register.

        Returns:
            Callable: The hook (not changed).
        """
        self.__hooks["on_iq"].append(hook)
        return hook
    
    def hook_send_message(self, hook:Callable) -> Callable:
        """
        Registers a hook for the send message event.
        The hook will be called when the client sends a message.
        
        Args:
            hook (Callable): The hook to register.
        
        Returns:
            Callable: The hook (not changed).
        """
        self.__hooks["send_message"].append(hook)
        return hook


    def _recv_xml(self) -> XmlElement:
        data = self.socket.recv(4096)
        return XmlParser.parse_elements(data.decode("utf-8"))[0]
    
    def _send_xml(self, xml:XmlElement):
        self.socket.sendall(xml.to_string().encode("utf-8"))


    def _start_xmpp_stream(self):
        logger.debug(f"Starting XMPP stream...")

        stream_start = XmlElement(
            "stream:stream", 
            attributes = [
                XmlAttribute("xmlns", "jabber:client"), 
                XmlAttribute("xmlns:stream", "http://etherx.jabber.org/streams"), 
                XmlAttribute("version", "1.0"), 
                XmlAttribute("to", self.host)
            ],
            is_closed=False
        )

        self._send_xml(stream_start)
    
    def _close_xmpp_stream(self):
        logger.debug(f"Closing XMPP stream...")

        self.socket.sendall(b"</stream:stream>")
    
    def _send_presence(self):
        logger.debug(f"Sending presence...")

        presence = XmlElement(
            "presence",
        )

        self._send_xml(presence)
    

    def _listen(self):
        logger.debug(f"Listening for XMPP stanzas...")

        buffer = ""

        while True:
            data = self.socket.recv(4096)
            if not data:
                self.disconnect()
                break

            buffer += data.decode("utf-8")

            try:
                elements = XmlParser.parse_elements(buffer)
                buffer = ""
            except Exception:
                # Incomplete stanza
                continue

            for element in elements:
                if element.name == "message":
                    message = XmppMessage(element)
                    hooks_result = self._trigger_hooks("on_message", message)
                    if hooks_result is None:
                        continue
                    self._trigger_handlers("message", hooks_result)

                elif element.name == "presence":
                    hooks_result = self._trigger_hooks("on_presence", element)
                    if hooks_result is None:
                        continue
                    self._trigger_handlers("presence", hooks_result)
                
                elif element.name == "iq":
                    hooks_result = self._trigger_hooks("on_iq", element)
                    if hooks_result is None:
                        continue
                    self._trigger_handlers("iq", hooks_result)

    def connect_feature(self, feature:XmppFeature, permissions: List[XmppPermission] | XmppPermission.ALL) -> None:
        """
        Connects the given feature to the XMPP client.

        Args:
            feature (XmppFeature): The feature to connect.
            permissions (List[XmppPermission] | XmppPermission.ALL): The permissions to grant.
        
        Example:
            >>> client.connect_feature(BindFeature("osmxmpp"), XmppPermission.ALL)
        """
        
        logger.debug(f"Connecting feature '{feature.ID}'...")

        XmppValidation.validate_id(feature.ID)

        feature_ci = XmppClientInterface(self, feature, permissions)

        feature._connect_ci(feature_ci)
        self.__features[feature.ID] = feature_ci
        self.__features_queue.append(feature.ID)

    def connect_features(self, features_with_permissions: List[Tuple[XmppFeature, List[XmppPermission] | XmppPermission.ALL]] ) -> None:
        """
        Connects the given features to the XMPP client.

        Args:
            features_with_permissions (List[Tuple[XmppFeature, List[XmppPermission] | XmppPermission.ALL]]): The features with permissions to connect
        
        Example:
            >>> client.connect_features([
            ...     (TLSFeature(), [XmppPersmision.SEND_XML, XmppPersmision.RECV_XML])
            ...     (BindFeature("osmxmpp"), XmppPermission.ALL)
            ... ])
        """

        for feature_with_permissions in features_with_permissions:
            self.connect_feature(feature_with_permissions[0], feature_with_permissions[1]) 


    def connect_extension(self, extension:XmppExtension, permissions: List[XmppPermission] | XmppPermission.ALL) -> None:
        """
        Connects the given extension to the XMPP client.

        Args:
            extension (XmppExtension): The extension to connect.
            permissions (List[XmppPermission] | XmppPermission.ALL): The permissions to grant.
        
        Example:
            >>> client.connect_extension(SomeExtension(), XmppPermission.ALL)
        """

        logger.debug(f"Connecting extension '{extension.ID}'...")

        XmppValidation.validate_id(extension.ID)

        extension_ci = XmppClientInterface(self, extension, permissions)
        
        extension._connect_ci(extension_ci)

        self.__extensions[extension.ID] = extension_ci
        self.__extensions_queue.append(extension.ID)

        self.__extensions[extension.ID].object._process()
    
    def connect_extensions(self, extensions_with_permissions: List[Tuple[XmppExtension, List[XmppPermission] | XmppPermission.ALL]] ) -> None:
        """
        Connects the given extensions to the XMPP client.

        Args:
            extensions_with_permissions (List[Tuple[XmppExtension, List[XmppPermission] | XmppPermission.ALL]]): The extensions with permissions to connect
        
        Example:
            >>> client.connect_extensions([
            ...     (SomeExtension(), [XmppPersmision.SEND_XML, XmppPersmision.RECV_XML])
            ...     (SomeOtherExtension(), XmppPermission.ALL)
            ... ])
        """

        for extension_with_permissions in extensions_with_permissions:
            self.connect_extension(extension_with_permissions[0], extension_with_permissions[1])


    def connect(self) -> None:
        """
        Connects to the XMPP server.
        """
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as self.socket:
            self.socket.connect((self.host, self.port))

            self.__connected = True

            logger.info(f"Connected to {self.host}:{self.port}")

            self._trigger_handlers("connected")
            
            self._start_xmpp_stream()

            # Features can sometimes be placed immediately at the start of the stream
            def recv_xml_features():
                xml = self._recv_xml()
                name = "stream:features"

                if (xml.name == name):
                    return xml

                features = xml.children[0].get_child_by_name(name)
                if (features):
                    return features

                return self._recv_xml()
                

            features_xml = recv_xml_features()
            while True:
                if (features_xml is None):
                    raise Exception("No stream features received")
                    
                processed_feature = None
                for feature_id in self.__features_queue:
                    feature = self.__features[feature_id].object

                    feature_xml = features_xml.get_child_by_name(feature.TAG)
                    if feature_xml:
                        logger.debug(f"Processing feature '{feature.ID}'...")
                        
                        feature._process(feature_xml)
                        processed_feature = feature
                        break
                
                if (processed_feature and not processed_feature.RECEIVE_NEW_FEATURES):
                    break

                features_xml = recv_xml_features()
            
            self._send_presence()

            self._trigger_handlers("ready")

            self._listen()
            self.socket.close()
    
    def disconnect(self):
        """
        Disconnects from the XMPP server.
        """

        if not self.__connected:
            raise Exception("XmppClient is not connected")

        self._close_xmpp_stream()
        self.socket.close()
        self.__connected = False

        self._trigger_handlers("disconnected")
        logger.info(f"Disconnected from {self.host}:{self.port}")
    

    def __repr__(self):
        return f"<XmppClient {self.host}:{self.port}>"



from .abc import XmppFeature
from .tls import TlsFeature
from .sasl import SaslException, SaslMechanism, SaslFeature, PlainMechanism
from .bind import BindFeature

__all__ = [
    "XmppFeature",

    "TlsFeature",

    "SaslException",
    "SaslMechanism",
    "SaslFeature",
    "PlainMechanism",

    "BindFeature",
]


import uuid

from typing import List

from osmxml import XmlParser
from osmxml import XmlElement
from osmxml import XmlAttribute
from osmxml import XmlTextElement

from osmomemo import OmemoBundle


class OmemoXml:
    @staticmethod
    def send_presence(jid_to: str) -> XmlElement:
        xml_str = f"""
        <presence to='{jid_to}'>
          <show>chat</show>
          <status>Available for OMEMO</status>
        </presence>
        """
        return XmlParser.parse_elements(xml_str)[0]

    @staticmethod
    def send_subscribe(jid_to: str) -> XmlElement:
        xml_str = f"""
        <presence type='subscribe' to='{jid_to}'>
          <status>OMEMO setup - requesting subscription</status>
        </presence>
        """
        return XmlParser.parse_elements(xml_str)[0]

    @staticmethod
    def send_subscribed(jid_to: str) -> XmlElement:
        xml_str = f"""
        <presence type='subscribed' to='{jid_to}'/>
        """
        return XmlParser.parse_elements(xml_str)[0]

    @staticmethod
    def check_node_exists(jid: str) -> XmlElement:
        xml_str = f"""
        <iq type='get' id='{OmemoXml.make_id()}' to='{jid}'>
          <query xmlns='http://jabber.org/protocol/disco#items' node='urn:xmpp:omemo:2:devices'/>
        </iq>
        """
        return XmlParser.parse_elements(xml_str)[0]

    @staticmethod
    def publish_device_list_setup(jid: str) -> XmlElement:
        xml_str = f"""
        <iq type='set' id='{OmemoXml.make_id()}' to='{jid}'>
          <pubsub xmlns='http://jabber.org/protocol/pubsub'>
            <create node='urn:xmpp:omemo:2:devices'/>
            <configure>
              <x xmlns='jabber:x:data' type='submit'>
                <field var='FORM_TYPE' type='hidden'>
                  <value>http://jabber.org/protocol/pubsub#node_config</value>
                </field>
                <field var='pubsub#access_model'><value>open</value></field>
                <field var='pubsub#persist_items'><value>true</value></field>
                <field var='pubsub#max_items'><value>1</value></field>
              </x>
            </configure>
          </pubsub>
        </iq>
        """
        return XmlParser.parse_elements(xml_str)[0]

    @staticmethod
    def publish_device(jid: str, device: int, lable: str = "OsmiumNet") -> XmlElement:
        xml_str = f"""
        <iq to='{jid}' type='set' id='{OmemoXml.make_id()}' >
          <pubsub xmlns='http://jabber.org/protocol/pubsub'>
            <publish node='urn:xmpp:omemo:2:devices'>
              <item id='current'>
                <devices xmlns='urn:xmpp:omemo:2'>
                  <device id='{device}' label='{lable}'/>
                </devices>
              </item>
            </publish>
          </pubsub>
        </iq>
        """
                
        return XmlParser.parse_elements(xml_str)[0]

    @staticmethod
    def publish_bundle_information(jid: str, bundle: OmemoBundle) -> XmlElement:
        device_id = bundle.get_device_id()
        ik = bundle.get_indentity().get_base64_public_key() 
        spk = bundle.get_prekey().get_base64_public_key()
        spk_sign = bundle.get_prekey_signature()

        opks_xml = XmlElement(
            name="prekeys",
            children=[
                XmlElement(
                    name="pk",
                    attributes=[
                        XmlAttribute("id", str(i))
                    ],
                    children=[
                        XmlTextElement(opk.get_base64_public_key())
                    ]
                ) for i, opk in bundle.get_onetime_prekeys().items()
            ]
        ).to_string()

        xml_str = f"""
        <iq from='{jid}' type='set' id='{OmemoXml.make_id()}'>
          <pubsub xmlns='http://jabber.org/protocol/pubsub'>
            <publish node='urn:xmpp:omemo:2:bundles'>
              <item id='{device_id}'>
                <bundle xmlns='urn:xmpp:omemo:2'>
                  <ik>{ik}</ik>
                  <spk id='0'>{spk}</spk>
                  <spks>{spk_sign}</spks>
                  <prekeys>
                    {opks_xml} 
                  </prekeys>
                </bundle>
              </item>
            </publish>
          </pubsub>
        </iq>
        """
        return XmlParser.parse_elements(xml_str)[0]

    @staticmethod
    def fetch_devices(jid: str, jid_to: str) -> XmlElement:
        xml_str = f"""
        <iq type='get'
            from='{jid}'
            to='{jid_to}'
            id='{OmemoXml.make_id()}'>
          <pubsub xmlns='http://jabber.org/protocol/pubsub'>
            <items node='urn:xmpp:omemo:2:devices'/>
          </pubsub>
        </iq>
        """
        return XmlParser.parse_elements(xml_str)[0]

    @staticmethod
    def fetch_bundles(jid: str, jid_to: str, device_to: int) -> XmlElement:
        xml_str = f"""
        <iq type='get' from='{jid}' to='{jid_to}' id='{OmemoXml.make_id()}'>
          <pubsub xmlns='http://jabber.org/protocol/pubsub'>
            <items node='urn:xmpp:omemo:2:bundles'>
              <item id='{device_to}'/>
            </items>
          </pubsub>
        </iq>
        """
        return XmlParser.parse_elements(xml_str)[0]

    @staticmethod
    def send_init_message(
                jid: str, 
                jid_to: str, 
                device: int, 
                device_to: int,
                key_data: str,
            ) -> XmlElement:
        xml_str = f"""
        <message from="{jid}" to="{jid_to}" type="chat" id="{OmemoXml.make_id()}">
            <encrypted xmlns="urn:xmpp:omemo:2">
                <header sid="{device}">
                    <keys jid="{jid_to}">
                        <key rid="{device_to}" kex="true">
                             {key_data}
                        </key>
                    </keys>
                </header>
                <payload>
                     {XmlTextElement("ciphertext-of-sce-envelope").to_string()}
                </payload>
            </encrypted>
            <body>
                [This message is OMEMO encrypted.]
            </body>
        </message>
        """
        return XmlParser.parse_elements(xml_str)[0]

    @staticmethod
    def send_message(
                jid: str, 
                jid_to: str, 
                device: int, 
                keys: List[XmlElement],
            ) -> XmlElement:
        xml_str_keys = "\n".join([nkey.to_string(raw=False) for nkey in keys]) 
        xml_str = f"""
        <message from="{jid}" to="{jid_to}" type="chat" id="{OmemoXml.make_id()}">
            <encrypted xmlns="urn:xmpp:omemo:2">
                <header sid="{device}">
                    {xml_str_keys}
                </header>
                <payload>
                     {XmlTextElement("ciphertext-of-sce-envelope").to_string()}
                </payload>
            </encrypted>
            <body>
                [This message is OMEMO encrypted.]
            </body>
        </message>
        """
        return XmlParser.parse_elements(xml_str)[0]

    @staticmethod
    def make_id():
        return str(uuid.uuid4()) 




import os
import base64
import json
import time
import random
import struct
import secrets

from typing import Callable, List, Tuple, Dict

from osmxml import XmlParser, XmlElement, XmlAttribute, XmlTextElement

from osmomemo import Omemo, OmemoBundle, XKeyPair, EdKeyPair
from osmomemo.storage import OmemoStorage

from ..abc import XmppExtension
from ...message import XmppMessage
from ...permission import XmppPermission

from .xml import OmemoXml


class OmemoExtension(XmppExtension):
    """
    XEP-0384: OMEMO Encryption implementation.
    """

    ID = "osmiumnet.omemo"

    # List of required permissions
    REQUIRED_PERMISSIONS: List[XmppPermission] = [
        XmppPermission.GET_JID,
        XmppPermission.SEND_XML,
        XmppPermission.LISTEN_ON_READY,
        XmppPermission.LISTEN_ON_IQ,
        XmppPermission.HOOK_ON_MESSAGE,
        XmppPermission.HOOK_SEND_MESSAGE,
    ]

    def __init__(self, bundle: OmemoBundle, storage: OmemoStorage):
        self.__bundle = bundle
        self.__omemo = Omemo(self.__bundle, storage)

        self.__registered_xmls: Dict[str, str] = {}

        self.__contact_bundles = {}

    def _connect_ci(self, ci):
        self.__ci = ci

    def _process(self):
        # Listeners
        @self.__ci.on_ready
        def on_ready():
            self.__on_ready()

        @self.__ci.on_iq
        def on_iq(iq):
            self.__on_iq(iq)


        # Hooks
        @self.__ci.hook_on_message
        def hook_on_message(message: XmppMessage):
            return self.__hook_on_message(message)

        @self.__ci.hook_send_message
        def hook_send_message(message: XmppMessage, *args, **kwargs):
            return self.__hook_send_message(message)


        # Variables
        self.__ci.variables.function(self.publish_bundle_information)

        self.__ci.variables.function(self.fetch_bundles)

    def publish_bundle_information(self):
        """
        Publishes the bundle information.

        Example:
            >>> client.extensions["osmiumnet.omemo"].publish_bundle_information()
        """

        xml = OmemoXml.publish_device(self.__ci.get_jid(False), self.__bundle.get_device_id())
        self.__send_registered_xml(xml, "publish_device:func")

    def fetch_bundles(self, jid):
        """
        Fetches the bundles from the given JID.

        Args:
            jid (str | List[str]): The JID(s) to fetch the bundles from.
        
        Example:
            >>> client.extensions["osmiumnet.omemo"].fetch_bundles("john@jabber.org")
        """

        def _fetch(jid: str):
            xml = OmemoXml.fetch_devices(self.__ci.get_jid(), jid)
            self.__send_registered_xml(xml, "fetch_devices:func")

        if (isinstance(jid, list)):
            for j in jid:
                _fetch(j)
        else:
            _fetch(jid)

    def __send_registered_xml(self, xml: XmlElement, name: str):
        self.__register_xml(xml, name)
        self.__ci.send_xml(xml)


    # Manage text to ensure they were sent from the extension 
    def __register_xml(self, xml: XmlElement, name: str):
        xml_id = xml.get_attribute_by_name("id").value
        if (xml_id):
            self.__registered_xmls[xml_id] = name

    def __get_xml_registration(self, xml: XmlElement) -> Tuple[bool, str]:
        is_registered = False
        name = ""

        id_attr = xml.get_attribute_by_name("id")
        if (id_attr):
            xml_id =    id_attr.value 
            is_registered = xml_id in self.__registered_xmls 
            if (is_registered):
                name = self.__registered_xmls[xml_id]
                del self.__registered_xmls[xml_id]
        return is_registered, name
    


    # Client events 
    def __on_ready(self):
        pass

    def __on_iq(self, iq):
        is_registered, name = self.__get_xml_registration(iq)

        if (not is_registered):
            return

        if (name == "publish_device:func"):
            xml = OmemoXml.publish_bundle_information(self.__ci.get_jid(False), self.__bundle)
            self.__send_registered_xml(xml, "publish_bundle_information")
        elif (name == "fetch_devices:func"):
            self.__parse_devices_response(iq)

            # Fetch bundles for contacts devices that are not cached yet
            for contact_jid, devices in self.__contact_bundles.items():
                for contact_device, bundle in self.__contact_bundles[contact_jid].items():
                    if (bundle == {}):
                        xml = OmemoXml.fetch_bundles(self.__ci.get_jid(), contact_jid, contact_device)
                        self.__send_registered_xml(xml, "fetch_bundles")
        elif (name == "fetch_bundles"):
            self.__parse_bundle_response(iq)

    def __hook_on_message(self, message: XmppMessage):
        final_message = None

        jid_from = message.from_jid.split("/")[0]
        if (message.encrypted):
            # Find and store JID keys
            device_keys = None
            for jid_keys in message.encrypted.header.xml.children:
                jid = jid_keys.get_attribute_by_name("jid").value
                if (self.__ci.get_jid(False) == jid):
                    device_keys = jid_keys
                    break

            if (not device_keys):
                return message

            # Find and store that device key
            device_key = None
            for key in device_keys.children:
                rid = key.get_attribute_by_name("rid").value
                device_rid = self.__bundle.get_device_id()
                if (device_rid == int(rid)):
                    device_key = key
                    break

            if (not device_key):
                return message

            key_data = device_key.children[0].to_string()
            device_from = int(message.encrypted.header.xml.get_attribute_by_name("sid").value)

            devices = self.__omemo.get_device_list(jid_from)
            if (type(devices) == list):
                key_data_js = json.loads(base64.b64decode(key_data).decode('utf-8'))

                wrapped = key_data_js["k"]
                payload = key_data_js["p"]

                wrapped_key_bytes = base64.b64decode(wrapped.encode("utf-8"))
                payload_bytes = base64.b64decode(payload.encode("utf-8"))

                de_message = self.__omemo.receive_message(
                            jid=jid_from, 
                            device=device_from,
                            wrapped_message_key=wrapped_key_bytes, 
                            payload=payload_bytes
                )

                final_massage = message
                final_massage.body = de_message.decode("utf-8")
            else:
                de_message = self.__receive_init_message(jid_from, device_from, key_data)

                if (de_message):
                    final_massage = message
                    final_massage.body = de_message

        return final_message if final_message else message 

    def __hook_send_message(self, message: XmppMessage):
        final_message = None

        jid_to = message.to_jid.split("/")[0]
        body = message.body
        encrypted_message = None

        devices = self.__omemo.get_device_list(jid_to)
        if (type(devices) == list):
            xml_keys = XmlElement("keys", [XmlAttribute("jid", jid_to)])
            payload = ""
            for device in devices:
                wrapped_key_bytes, payload_bytes = self.__omemo.send_message(
                            jid=jid_to,
                            device=device,
                            message_bytes=body.encode("utf-8")
                )

                wrapped = base64.b64encode(wrapped_key_bytes).decode("utf-8")
                payload = base64.b64encode(payload_bytes).decode("utf-8")

                key_data_blob = json.dumps({
                    "k": wrapped,
                    "p": payload
                }).encode('utf-8')

                key_data = base64.b64encode(key_data_blob).decode('utf-8')

                xml_keys.add_child(
                    XmlElement(
                        "key", 
                        [XmlAttribute("rid", device)],
                        [XmlTextElement(key_data)]
                    )
                )

            encrypted_message = OmemoXml.send_message(
                        self.__ci.get_jid(),
                        jid_to,
                        self.__bundle.get_device_id(),
                        [xml_keys]
            )
        elif (devices is None):
            encrypted_message = self.__send_init_message(jid_to, body)

        if (encrypted_message):
            if (not final_message):
                final_message = message
            # Wrap encrypted message into message
            final_message.xml.add_child(encrypted_message.get_child_by_name("encrypted"))
            final_message.xml.remove_child_by_name("body")
            final_message.xml.add_child(encrypted_message.get_child_by_name("body"))

        return final_message if final_message else message 

    # Parse device information from IQ response
    def __parse_devices_response(self, iq):
        try:
            # Check if this is a devices response
            pubsub = iq.get_child_by_name("pubsub")
            if (not pubsub):
                return
                
            items = pubsub.get_child_by_name("items")
            if (not items or items.get_attribute_by_name("node").value != "urn:xmpp:omemo:2:devices"):
                return
                
            item = items.get_child_by_name("item")
            if (not item):
                return

            devices = item.get_child_by_name("devices")
            if (not devices):
                return
            
            contact_jid = iq.get_attribute_by_name("from").value.split("/")[0]
            if contact_jid not in self.__contact_bundles:
                self.__contact_bundles[contact_jid] = {}

            for device in devices.children:
                device_id = int(device.get_attribute_by_name("id").value)
                self.__contact_bundles[contact_jid][device_id] = {}
        except Exception as e:
            print(f"Error parsing devices: {e}")
        
    # Parse bundle information from IQ response
    def __parse_bundle_response(self, iq):
        try:
            # Check if this is a bundle response
            pubsub = iq.get_child_by_name("pubsub")
            if (not pubsub):
                return
                
            items = pubsub.get_child_by_name("items")
            if (not items or items.get_attribute_by_name("node").value != "urn:xmpp:omemo:2:bundles"):
                return
                
            item = items.get_child_by_name("item")
            if (not item):
                return
                
            bundle = item.get_child_by_name("bundle")
            if (not bundle):
                return
                
            # Extract bundle information
            device_id = int(item.get_attribute_by_name("id").value)
            contact_jid = iq.get_attribute_by_name("from").value.split("/")[0]
            
            bundle_data = {
                "spk": bundle.get_child_by_name("spk").children[0].to_string().strip(),
                "spks": bundle.get_child_by_name("spks").children[0].to_string().strip(), 
                "ik": bundle.get_child_by_name("ik").children[0].to_string().strip(),
                "opks": {}
            }
            
            prekeys_elem = bundle.get_child_by_name("prekeys")
            if prekeys_elem:
                for pk in prekeys_elem.children:
                    pk_id = pk.get_attribute_by_name("id").value
                    pk_data = pk.children[0].to_string()
                    bundle_data["opks"][pk_id] = pk_data
            
            # Store the bundle
            if contact_jid not in self.__contact_bundles:
                self.__contact_bundles[contact_jid] = {}
            self.__contact_bundles[contact_jid][device_id] = bundle_data
        except Exception as e:
            print(f"Error parsing bundle: {e}")


    def __send_init_message(self, jid_to: str, message: str):
        if (jid_to in self.__contact_bundles):
            # TODO: many devices
            device_to = list(self.__contact_bundles[jid_to].keys())[0]
            bundle_to = self.__contact_bundles[jid_to][device_to]
            
            # Choose random opk id
            opk_id = random.choice(list(bundle_to["opks"].keys())) 

            ek_pub, en_message = self.__omemo.create_init_message(
                    jid=jid_to,
                    device=device_to,
                    message_bytes=message.encode("utf-8"),
                    indentity_key=EdKeyPair.base64_to_public_key(bundle_to["ik"]),
                    signed_prekey=XKeyPair.base64_to_public_key(bundle_to["spk"]),
                    prekey_signature=base64.b64decode(bundle_to["spks"].encode("utf-8")),
                    onetime_prekey=XKeyPair.base64_to_public_key(bundle_to["opks"][opk_id])
            )
            
            key_data_blob = json.dumps({
                "ik": self.__bundle.get_indentity().get_base64_public_key(),
                "ek": XKeyPair.public_key_to_base64(ek_pub),
                "spk_id": "0",
                "opk_id": opk_id,
                "ct": base64.b64encode(en_message).decode("utf-8") 
            }).encode('utf-8')

            key_data = base64.b64encode(key_data_blob).decode('utf-8')

            xml_message = OmemoXml.send_init_message(
                    jid=self.__ci.get_jid(),
                    jid_to=jid_to,
                    device=self.__bundle.get_device_id(),
                    device_to=device_to,
                    key_data=key_data 
            )
            
            return xml_message

    def __receive_init_message(self, jid_from: str, device_from: int, wrapped: str):
        message = None
        if (jid_from in self.__contact_bundles):
            key_data_js = json.loads(base64.b64decode(wrapped).decode('utf-8'))

            en_message = base64.b64decode(key_data_js["ct"].encode("utf-8"))
            indentity_key = EdKeyPair.base64_to_public_key(key_data_js["ik"])
            ephemeral_key = XKeyPair.base64_to_public_key(key_data_js["ek"])
            spk_id = key_data_js["spk_id"]
            opk_id = key_data_js["opk_id"]

            de_message = self.__omemo.accept_init_message(
                    jid=jid_from,
                    device=device_from,
                    encrypted_message=en_message,
                    indentity_key=indentity_key,
                    ephemeral_key=ephemeral_key,
                    spk_id=spk_id,
                    opk_id=opk_id
            )

            message = de_message.decode("utf-8")

        return message




from typing import Callable, List, Tuple, Dict

from osmxml import XmlElement

from ...abc import XmppExtension
from ....message import XmppMessage
from ....permission import XmppPermission

from .xml import SubscriptionXml


class SubscriptionExtension(XmppExtension):
    """
    XEP-0379: Pre-Authenticated Roster Subscription implementation.
    """

    ID = "osmiumnet.roster.subscription"

    # List of required permissions
    REQUIRED_PERMISSIONS: List[XmppPermission] = [
        XmppPermission.GET_JID,
        XmppPermission.SEND_XML,
        XmppPermission.LISTEN_ON_READY,
        XmppPermission.LISTEN_ON_PRESENCE,
        XmppPermission.LISTEN_ON_IQ,
    ]

    def __init__(self):
        self.__ensure_list: List[str] = [] 

        self.__handlers = {
            "on_check_subscriptions": [],
        }
     
    def _connect_ci(self, ci):
        self.__ci = ci

    def _process(self):
        # Listeners
        @self.__ci.on_presence
        def on_presence(presence: XmlElement):
            self.__on_presence(presence)
        
        # Hooks
        @self.__ci.hook_on_iq
        def hook_on_iq(iq: XmlElement):
            return self.__hook_on_iq(iq)

        # Variables
        self.__ci.variables.function(self.on_check_subscriptions)

        self.__ci.variables.function(self.check_for_subscriptions)

        self.__ci.variables.function(self.ensure_subscription)

    def on_check_subscriptions(self, handler:Callable):
        """
        Registers a handler for the ``on_check_subscriptions`` event.

        Handler will be called when subscriptions are received.

        Args:
            handler (Callable): The handler to register.
        
        Returns:
            Callable: The handler (unchanged).
        
        Example:
            >>> @client.extensions["osmiumnet.roster.subscription"].on_check_subscriptions
            ... def on_check_subscriptions(iq):
            ...     print(f"Received subscriptions: {iq.body}")
        """

        self.__handlers["on_check_subscriptions"].append(handler)
        return handler

    def check_for_subscriptions(self):
        """
        Sends a subscription check request.

        Example:
            >>> client.extensions["osmiumnet.roster.subscription"].check_for_subscriptions()
        """

        xml = SubscriptionXml.check_for_subscription(self.__ci.get_jid(False))
        self.__ci.send_xml(xml)

    def ensure_subscription(self, jid_to: str):
        """
        Ensures that the JID is subscribed.

        Args:
            jid_to (str): The JID to ensure.

        Example:
            >>> client.extensions["osmiumnet.roster.subscription"].ensure_subscription("john@jabber.org")
        """
        self.__ensure_list.append(jid_to)


    def __on_presence(self, presence: XmlElement):
        if(SubscriptionXml.send_subscribe_filter(presence)):
            # If it receives subscribe request, and jid in ensure list,
            # it sends subscribed to it
            jid = presence.get_attribute_by_name("from")
            if (jid in self.__ensure_list):
                # Send subscribed
                xml = SubscriptionXml.send_subscribed(jid)
                self.__ci.send_xml(xml)

    def __hook_on_iq(self, iq: XmlElement):
        calls = [
            self.__process_check_subscriptions(iq),
        ]

        if True in calls:
            return

        return iq
    
    def __process_check_subscriptions(self, iq: XmlElement):
        iq_type = iq.get_attribute_by_name("type").value
        if not (SubscriptionXml.check_for_subscription_filter(iq) and iq_type == "result"):
            return False
        
        query = iq.get_child_by_name("query")
        
        # Get subscribed query children
        subscribed_children_list = []
        ask_children_list = []
        for query_child in query.children:
            if (query_child.name == "item"):
                jid = query_child.get_attribute_by_name("jid").value
                # Check subscription status
                subscription_status = query_child.get_attribute_by_name("subscription").value
                if (subscription_status == "both"):
                    subscribed_children_list.append(jid)
                elif (subscription_status == "none"):
                    ask =  query_child.get_attribute_by_name("ask").value
                    if (ask == "subscribe"):
                        ask_children_list.append(jid)

        # If jid from ensure list is not subsribed, it sends subscription to it
        for ensure_jid in self.__ensure_list:
            if (ensure_jid not in subscribed_children_list):
                # Send to subscribe
                xml = SubscriptionXml.send_subscribe(ensure_jid)
                self.__ci.send_xml(xml)

                xml = SubscriptionXml.send_presence(ensure_jid)
                self.__ci.send_xml(xml)

        # If jid from ask list is in ensure list, it sends subscribed to it
        for ask_child in ask_children_list:
            if (ask_child in self.__ensure_list):
                # Send subscribed
                xml = SubscriptionXml.send_subscribed(ensure_jid)
                self.__ci.send_xml(xml)

        # Trigger event
        handlers = self.__handlers["on_check_subscriptions"] 
        for handler in handlers:
            handler(iq)
        
        return True



from enum import Enum, auto


class XmppPermission(Enum):
    """
    Permissions are used to control what actions Feature or Extension can perform.
    """

    ALL = auto()

    SEND_XML = auto()
    RECV_XML = auto()

    GET_JID = auto()
    GET_RESOURCE = auto()
    SET_JID = auto()
    SET_RESOURCE = auto()

    GET_HOST = auto()
    GET_PORT = auto()

    GET_SOCKET = auto()
    CHANGE_SOCKET = auto()

    OPEN_STREAM = auto()

    LISTEN_ON_CONNECT = auto()
    LISTEN_ON_DISCONNECT = auto()
    LISTEN_ON_READY = auto()
    LISTEN_ON_MESSAGE = auto()
    LISTEN_ON_PRESENCE = auto()
    LISTEN_ON_IQ = auto()

    HOOK_ON_MESSAGE = auto()
    HOOK_ON_PRESENCE = auto()
    HOOK_ON_IQ = auto()

    HOOK_SEND_MESSAGE = auto()

    DISCONNECT = auto()

    def __str__(self):
        return self.name
    
    def __repr__(self):
        return f"<XmppPermission.{self.name}>"


import socket

from typing import List, Callable

from .validation import XmppValidation
from .permission import XmppPermission

from osmxml import *

import logging


logger = logging.getLogger(__name__)


class XmppClientInterface:
    """
    Xmpp client interface implementation.
    Used by Features or Extensions to interact with the XMPP client.
    """

    def __init__(self, client, obj, permissions: List[XmppPermission] | XmppPermission.ALL):
        """
        Initializes the Xmpp client interface.

        Args:
            client (XmppClient): The XMPP client.
            permissions (List[XmppPermission] | XmppPermission.ALL): The permissions to grant.
        """

        self.__client = client
        self.variables = XmppVariableInterface(self)
        
        if permissions == XmppPermission.ALL:
            permissions = [XmppPermission.ALL]

        self.__permissions = permissions

        self.object = obj
    
    def __handle_permission(self, permission:XmppPermission):
        if self.has_permission(permission):
            return
        
        raise Exception(f"No {permission} permission")
    
    # Exposed functions
    def has_permission(self, *permissions) -> bool:
        """
        Checks if the client interface has the given permission.

        Args:
            permissions (XmppPermission): The permissions to check.

        Returns:
            bool: True if the client interface has the permission, False otherwise.
        """
        if XmppPermission.ALL in self.__permissions:
            return True
        
        for permission in permissions:
            if permission in self.__permissions:
                return True

        return False

    def send_xml(self, xml:XmlElement):
        """
        Sends an XML element to the XMPP client.
        Requires the SEND_XML permission.

        Args:
            xml (XmlElement): The XML element to send.

        Returns:
            XmlElement: The XML element sent.
        """
        self.__handle_permission(XmppPermission.SEND_XML)
        return self.__client._send_xml(xml)
    
    def recv_xml(self) -> XmlElement:
        """
        Receives an XML element from the XMPP client.
        Requires the RECV_XML permission.

        Returns:
            XmlElement: The XML element received.
        """
        self.__handle_permission(XmppPermission.RECV_XML)
        return self.__client._recv_xml()
    
    def get_jid(self, with_resouce:bool = True) -> str:
        """
        Gets the JID of the XMPP client.
        Requires the GET_JID permission.

        Args:
            with_resouce (bool): Whether to include the resource in the JID. (Default: True)

        Returns:
            str: The JID of the XMPP client.
        """
        self.__handle_permission(XmppPermission.GET_JID)

        if with_resouce:
            return self.__client.jid
        else:
            return self.__client.jid.split("/")[0]

    def get_resource(self) -> str:
        """
        Gets the resource of the XMPP client.
        Requires the GET_RESOURCE or GET_JID permissions.

        Returns:
            str: The resource of the XMPP client.
        """
        self.__handle_permission(XmppPermission.GET_RESOURCE, XmppPermission.GET_JID)
        return self.__client.resource

    def set_jid(self, jid:str):
        """
        Sets the JID of the XMPP client.
        Requires the SET_JID permission.

        Args:
            jid (str): The new JID of the XMPP client.
        """

        XmppValidation.validate_jid(jid)

        self.__handle_permission(XmppPermission.SET_JID)
        self.__client.jid = jid
        return

    def set_resource(self, resource:str):
        """
        Sets the resource of the XMPP client.
        Requires the SET_RESOURCE permission.

        Args:
            resource (str): The new resource of the XMPP client.
        """

        XmppValidation.validate_resource(resource)

        self.__handle_permission(XmppPermission.SET_RESOURCE)
        self.__client.resource = resource
        return

    def get_host(self) -> str:
        """
        Gets the host of the XMPP client.
        Requires the GET_HOST permission.

        Returns:
            str: The host of the XMPP client.
        """
        self.__handle_permission(XmppPermission.GET_HOST)
        return self.__client.host

    def get_port(self) -> int:
        """
        Gets the port of the XMPP client.
        Requires the GET_PORT permission.

        Returns:
            int: The port of the XMPP client.
        """
        self.__handle_permission(XmppPermission.GET_PORT)
        return self.__client.port
    
    def change_socket(self, socket):
        """
        Changes the socket of the XMPP client.
        Requires the CHANGE_SOCKET permission.

        Args:
            socket (socket): The new socket of the XMPP client.
        """
        self.__handle_permission(XmppPermission.CHANGE_SOCKET)
        self.__client.socket = socket
        return
    
    def get_socket(self) -> socket:
        """
        Gets the socket of the XMPP client.
        Requires the GET_SOCKET permission.

        Returns:
            socket: The socket of the XMPP client.
        """
        self.__handle_permission(XmppPermission.GET_SOCKET)
        return self.__client.socket
    
    def open_stream(self):
        """
        Opens the XMPP stream.
        Requires the OPEN_STREAM permission.
        """
        self.__handle_permission(XmppPermission.OPEN_STREAM)
        return self.__client._start_xmpp_stream()
    
    def on_connect(self, handler:Callable) -> Callable:
        """
        Registers a handler for the connected event.
        The handler will be called when the client is connected to the XMPP server, but not ready yet.

        Args:
            handler (Callable): The handler to register.

        Returns:
            Callable: The handler (not changed).
        """
        self.__handle_permission(XmppPermission.LISTEN_ON_CONNECT)
        return self.__client.on_connect(handler)
    
    def on_disconnect(self, handler:Callable) -> Callable:
        """
        Registers a handler for the disconnected event.
        The handler will be called when the client is disconnected from the XMPP server.

        Args:
            handler (Callable): The handler to register.

        Returns:
            Callable: The handler (not changed).
        """
        self.__handle_permission(XmppPermission.LISTEN_ON_DISCONNECT)
        return self.__client.on_disconnect(handler)
    
    def on_ready(self, handler:Callable) -> Callable:
        """
        Registers a handler for the ready event.
        The handler will be called when the client is ready to send and receive XMPP stanzas.

        Args:
            handler (Callable): The handler to register.

        Returns:
            Callable: The handler (not changed).
        """
        self.__handle_permission(XmppPermission.LISTEN_ON_READY)
        return self.__client.on_ready(handler)

    def on_message(self, handler:Callable) -> Callable:
        """
        Registers a handler for the message event.
        The handler will be called when the client receives a message stanza.

        Args:
            handler (Callable): The handler to register.

        Returns:
            Callable: The handler (not changed).
        """
        self.__handle_permission(XmppPermission.LISTEN_ON_MESSAGE)
        return self.__client.on_message(handler)
    
    def on_presence(self, handler:Callable) -> Callable:
        """
        Registers a handler for the presence event.
        The handler will be called when the client receives a presence stanza.

        Args:
            handler (Callable): The handler to register.

        Returns:
            Callable: The handler (not changed).
        """
        self.__handle_permission(XmppPermission.LISTEN_ON_PRESENCE)
        return self.__client.on_presence(handler)
    
    def on_iq(self, handler:Callable) -> Callable:
        """
        Registers a handler for the iq event.
        The handler will be called when the client receives an iq stanza.

        Args:
            handler (Callable): The handler to register.

        Returns:
            Callable: The handler (not changed).
        """
        self.__handle_permission(XmppPermission.LISTEN_ON_IQ)
        return self.__client.on_iq(handler)
    
    def hook_on_message(self, hook:Callable) -> Callable:
        """
        Registers a hook for the message event.
        The hook will be called when the client receives a message stanza.

        Args:
            hook (Callable): The hook to register.

        Returns:
            Callable: The hook (not changed).
        """
        self.__handle_permission(XmppPermission.HOOK_ON_MESSAGE)
        return self.__client.hook_on_message(hook)
    
    def hook_on_presence(self, hook:Callable) -> Callable:
        """
        Registers a hook for the presence event.
        The hook will be called when the client receives a presence stanza.

        Args:
            hook (Callable): The hook to register.

        Returns:
            Callable: The hook (not changed).
        """
        self.__handle_permission(XmppPermission.HOOK_ON_PRESENCE)
        return self.__client.hook_on_presence(hook)
    
    def hook_on_iq(self, hook:Callable) -> Callable:
        """
        Registers a hook for the iq event.
        The hook will be called when the client receives an iq stanza.

        Args:
            hook (Callable): The hook to register.

        Returns:
            Callable: The hook (not changed).
        """
        self.__handle_permission(XmppPermission.HOOK_ON_IQ)
        return self.__client.hook_on_iq(hook)
    
    def hook_send_message(self, hook:Callable) -> Callable:
        """
        Registers a hook for the send message event.
        The hook will be called when the client sends a message.
        
        Args:
            hook (Callable): The hook to register.
        
        Returns:
            Callable: The hook (not changed).
        """
        self.__handle_permission(XmppPermission.HOOK_SEND_MESSAGE)
        return self.__client.hook_send_message(hook)
    
    def disconnect(self):
        """
        Disconnects from the XMPP server.
        Requires the DISCONNECT permission.
        """
        self.__handle_permission(XmppPermission.DISCONNECT)
        return self.__client.disconnect()
    

    def __repr__(self):
        return f"<XmppClientInterface of '{repr(self.__client)}'>"

class XmppVariableInterface:
    """
    Xmpp variable interface implementation.
    Used by Features or Extensions to expose variables
    """

    def __init__(self, ci):
        """
        Initializes the XMPP variable interface.

        Args:
            client (XmppClient): The XMPP client.
        """

        super().__setattr__("ci", ci)
        super().__setattr__("variables", {})
    
    def function(self, function:Callable):
        """
        Registers a function to the variable.

        Args:
            function (Callable): The function to register.
        
        Returns:
            Callable: The function (not changed).
        """

        logging.debug(f"Registering function '{function.__name__}' for ''{super().__getattribute__("ci").object.ID}''...")

        super().__getattribute__("variables")[function.__name__] = function
        return function
    
    def __getattr__(self, name):
        if name in super().__getattribute__("variables"):
            return super().__getattribute__("variables")[name]
        else:
            raise AttributeError(f"No variable named '{name}'")
    
    def __setattr__(self, name, value):
        super().__getattribute__("variables")[name] = value



from typing import Callable, List, Tuple, Dict

from ...abc import XmppExtension
from ....message import XmppMessage
from ....permission import XmppPermission

from .xml import DiscoveryXml


class ServiceDiscoveryExtension(XmppExtension):
    """
    XEP-0030: Service Discovery implementation.
    """

    ID = "osmiumnet.service.discovery"

    # List of required permissions
    REQUIRED_PERMISSIONS: List[XmppPermission] = [
        XmppPermission.SEND_XML,
    ]

    def __init__(self):
        pass
     
    def _connect_ci(self, ci):
        self.__ci = ci
    
    def discover(self):
        """
        Sends a service discovery request.

        Example:
            >>> client.extensions["osmiumnet.service.discovery"].discover()
        """

        xml = DiscoveryXml.discover()
        self.__ci.send_xml(xml)

    def _process(self):
        # Variables
        self.__ci.variables.function(self.discover)




from osmxml import *


class _XmppMessageElement:
    def __init__(self, xml:XmlElement):
        self.__xml = xml
    
    @property
    def xml(self):
        return self.__xml

    def get_attribute_by_index(self, index:int):
        return self.__xml.get_attribute_by_index(index)
    
    def get_child_by_index(self, index:int):
        return _get_xmpp_message_element_or_text(self.__xml.get_child_by_index(index))
    

    def __getattr__(self, name):
        if self.__xml.get_attribute_by_name(name):
            return self.__xml.get_attribute_by_name(name).value
        
        return _get_xmpp_message_element_or_text(self.__xml.get_child_by_name(name))
    
    def __getitem__(self, index):
        return self.get_child_by_index(index)
    
    def __repr__(self):
        return f"<_XmppMessageElement {self.__xml.to_string()}>"

def _get_xmpp_message_element_or_text(xml: XmlElement) -> _XmppMessageElement | str:
    if xml == None:
        return

    has_one_child = len(xml.children) == 1

    if has_one_child and hasattr(xml.children[0], "text"):
        return xml.children[0].text
    else:
        return _XmppMessageElement(xml)
        

class XmppMessage:
    """
    XMPP message implementation.
    """

    def __init__(self, xml: XmlElement=None):
        if xml:
            self.__xml = xml
        else:
            self.__xml = XmlElement("message")

    @property
    def xml(self):
        """
        Gets the editable XML element of this message.
        """
        
        return self.__xml

    def get_attribute_by_index(self, index:int):
        return self.__xml.get_attribute_by_index(index)
    
    def get_child_by_index(self, index:int):
        return _get_xmpp_message_element_or_text(self.__xml.get_child_by_index(index))


    def __getattr__(self, name):
        if name == "from_jid":
            return self.__xml.get_attribute_by_name("from").value
        
        if name == "to_jid":
            return self.__xml.get_attribute_by_name("to").value
        
        if self.__xml.get_attribute_by_name(name):
            return self.__xml.get_attribute_by_name(name).value

        return _get_xmpp_message_element_or_text(self.__xml.get_child_by_name(name))
    
    def __getitem__(self, index):
        return self.get_child_by_index(index)
    
    def __repr__(self):
        return f"<XmppMessage from='{self.from_jid}' to='{self.to_jid}' type='{self.type}'>"



import uuid

from osmxml import XmlParser
from osmxml import XmlElement
from osmxml import XmlAttribute
from osmxml import XmlTextElement


class DiscoveryXml:
    @staticmethod
    def discover() -> XmlElement:
        xml_str = f"""
        <iq type='get' id='{DiscoveryXml.make_id()}'>
          <query xmlns='http://jabber.org/protocol/disco#info'/>
        </iq>
        """
        return XmlParser.parse_elements(xml_str)[0]

    @staticmethod
    def make_id():
        return str(uuid.uuid4()) 




import socket
import ssl

from typing import List

from .abc import XmppFeature
from ..permission import XmppPermission

from osmxml import *

import logging


logger = logging.getLogger(__name__)


class TlsFeature(XmppFeature):
    """
    TLS feature implementation.

    Attributes:
        ssl_context (ssl.SSLContext): The SSL context to use.
        verify_locations (List[str]): The locations to verify the server certificate.
    """

    ID = "osmiumnet.tls"
    TAG = "starttls"

    RECEIVE_NEW_FEATURES = True

    REQUIRED_PERMISSIONS: List[XmppPermission] = [
        XmppPermission.SEND_XML,
        XmppPermission.RECV_XML,
        XmppPermission.OPEN_STREAM,
        XmppPermission.GET_HOST,
        XmppPermission.GET_SOCKET,
        XmppPermission.CHANGE_SOCKET,
    ]

    def __init__(self, ssl_context=None, verify_locations=None):
        if ssl_context is None:
            logger.debug(f"Creating default SSL context...")
            self.__ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
            self.__ssl_context.check_hostname = True
            self.__ssl_context.verify_mode = ssl.CERT_REQUIRED

        if verify_locations is not None:
            logger.debug(f"Loading verify locations...")
            self.__ssl_context.load_verify_locations(verify_locations)
    
    def _connect_ci(self, ci):
        self.__ci = ci

    def _process(self, element):
        tls_handshake = XmlElement(
            "starttls",

            attributes = [
                XmlAttribute("xmlns", "urn:ietf:params:xml:ns:xmpp-tls")
            ]
        )

        logger.debug(f"Sending TLS handshake...")
        self.__ci.send_xml(tls_handshake)
        data = self.__ci.recv_xml()

        if data.name != "proceed":
            return None
        
        logger.debug(f"Wrapping socket...")
        tls_socket = self.__ssl_context.wrap_socket(self.__ci.get_socket(), server_hostname=self.__ci.get_host())
        
        logger.debug(f"Performing TLS handshake...")
        tls_socket.do_handshake()

        logger.debug(f"Done! Changing client socket...")
        self.__ci.change_socket(tls_socket)
        
        self.__ci.open_stream()



from .abc import XmppExtension
from .omemo import OmemoExtension 
from .service.discovery import ServiceDiscoveryExtension


__all__ = [
    "XmppExtension",
    "OmemoExtension",
    "ServiceDiscoveryExtension",
]



from abc import ABC, abstractmethod


class XmppFeature(ABC):
    """
    Features are used to implement specific XMPP stream features.

    Attributes:
        ID (str): The ID of the feature implementation.
        TAG (str): The tag of the feature. This is used to identify the feature in the XML stream.
        RECEIVE_NEW_FEATURES (bool): Whether the feature should receive new features.
    """

    ID = None
    TAG = None

    RECEIVE_NEW_FEATURES = None
    
    @abstractmethod
    def _connect_ci(self, ci) -> None:
        """
        Connects the feature to the client interface.

        Args:
            ci (XmppClientInterface): The client interface.
        """
        ...
    
    @abstractmethod
    def _process(self, element) -> None:
        """
        Processes the feature.

        Args:
            element (XmlElement): The XML element to process.
        """
        ...



import base64
from abc import ABC, abstractmethod

from typing import List

from .abc import XmppFeature
from ..permission import XmppPermission

from osmxml import *

import logging


logger = logging.getLogger(__name__)


class SaslException(Exception):
    pass

class SaslMechanism(ABC):
    """
    SASL mechanisms are used to authenticate the user.

    Attributes:
        NAME (str): The name of the mechanism.
    """

    NAME = None

    @abstractmethod
    def process(self, ci):
        """
        Processes the mechanism.

        Args:
            ci (XmppClientInterface): The client interface given from the SaslFeature.
        """
        ...


class PlainMechanism(SaslMechanism):
    """
    PLAIN SASL mechanism implementation.

    Attributes:
        username (str): The username to authenticate with.
        password (str): The password to authenticate with.
    """

    NAME = "PLAIN"

    def __init__(self, username:str, password:str):
        self.__auth_string = f"\0{username}\0{password}"
        self.__auth_string = base64.b64encode(self.__auth_string.encode("utf-8")).decode()

    def process(self, ci):
        auth_xml = XmlElement(
            "auth",

            attributes = [
                XmlAttribute("xmlns", "urn:ietf:params:xml:ns:xmpp-sasl"),
                XmlAttribute("mechanism", "PLAIN"),
            ],

            children = [
                XmlTextElement(self.__auth_string),
            ]
        )

        ci.send_xml(auth_xml)
        data = ci.recv_xml()

        return data


class SaslFeature(XmppFeature):
    """
    SASL feature implementation.

    Attributes:
        mechanisms (List[SaslMechanism]): The SASL mechanisms to use.
    
    Raises:
        SaslException: If authentication fails.
    """

    ID = "osmiumnet.sasl"
    TAG = "mechanisms"

    RECEIVE_NEW_FEATURES = True

    REQUIRED_PERMISSIONS: List[XmppPermission] = [
        XmppPermission.SEND_XML,
        XmppPermission.RECV_XML,
        XmppPermission.OPEN_STREAM,
    ]

    def __init__(self, mechanisms:List[SaslMechanism]):
        self.__mechanisms = mechanisms
    
    def _connect_ci(self, ci):
        self.__ci = ci
    
    def _process(self, element):
        mechanisms_xml = element

        logger.debug(f"Processing mechanisms...")

        recv_data = None
        for mechanism in self.__mechanisms:
            if mechanism.NAME in [mechanism_xml.children[0].text for mechanism_xml in mechanisms_xml.children]:
                logger.debug(f"Processing mechanism '{mechanism.NAME}'...")
                recv_data = mechanism.process(self.__ci)
                break
        
        if recv_data is None or recv_data.name != "success":
            raise SaslException("SASL authentication failed")
        
        self.__ci.open_stream()



import re


osmxmpp_id_regex = re.compile(r'^[A-Za-z0-9]+(?:\.[A-Za-z0-9]+)+$')

xmpp_jid_regex = re.compile(r'^(?:([a-zA-Z0-9._%+-]+)@)?([a-zA-Z0-9.-]+\.[a-zA-Z]{2,})(?:/(.+))?$')
xmpp_resource_regex = re.compile(r'^[^\s/]+$')


class ValidationException(Exception):
    pass


class XmppValidation:
    """
    OsmXmpp validation methods
    """

    @staticmethod
    def validate_id(id:str):
        """
        Validates the extension/feature ID.

        Args:
            id (str): The ID to validate.
        
        Raises:
            ValidationException: If the ID is invalid.
        """

        if type(id) != str:
            raise ValidationException(f"ID '{id}' is not a string")
        
        if not osmxmpp_id_regex.match(id):
            raise ValidationException(f"Invalid ID '{id}'")

    @staticmethod
    def validate_jid(jid:str):
        """
        Validates the JID.

        Args:
            jid (str): The JID to validate.

        Raises:
            ValidationException: If the JID is invalid.
        """

        if type(jid) != str:
            raise ValidationException(f"JID '{jid}' is not a string")
        
        if not xmpp_jid_regex.match(jid):
            raise ValidationException(f"Invalid JID '{jid}'")

    @staticmethod
    def validate_resource(resource:str):
        """
        Validates the resource.

        Args:
            resource (str): The resource to validate.

        Raises:
            ValidationException: If the resource is invalid.
        """

        if type(resource) != str:
            raise ValidationException(f"Resource '{resource}' is not a string")

        if not xmpp_resource_regex.match(resource):
            raise ValidationException(f"Invalid resource '{resource}'")



from setuptools import setup, find_packages

setup(
    name="osmxmpp",
    version="1.0.0",

    author="osmiumnet",
    description="Python XMPP library",

    packages=find_packages(),
    install_requires=[
        "osmxml==1.0.1",
        "osmomemo==1.1.2",
    ],
    python_requires=">=3.10",
)



__version__ = "0.1.0"
__author__ = "osmiumnet"

from .validation import XmppValidation, ValidationException

from .permission import XmppPermission

from .message import XmppMessage

from .ci import XmppClientInterface
from .client import XmppClient

from .extensions.abc import XmppExtension
from .extensions.omemo import OmemoExtension
from .extensions.service.discovery import ServiceDiscoveryExtension

from .features.abc import XmppFeature
from .features.tls import TlsFeature
from .features.sasl import SaslException, SaslMechanism, SaslFeature, PlainMechanism
from .features.bind import BindFeature

__all__ = [
    "XmppValidation",
    "ValidationException",

    "XmppMessage",

    "XmppPermission",

    "XmppClient",
    "XmppClientInterface",


    "XmppExtension",

    "OmemoExtension",
    
    "ServiceDiscoveryExtension",


    "XmppFeature",

    "TlsFeature",

    "SaslException",
    "SaslMechanism",
    "SaslFeature",
    "PlainMechanism",

    "BindFeature",
]



from typing import List

from .abc import XmppFeature
from ..permission import XmppPermission

from osmxml import *

import logging


logger = logging.getLogger(__name__)


class BindFeature(XmppFeature):
    """
    Bind feature implementation.

    Attributes:
        resource (str): The resource to bind to.
    """

    ID = "osmiumnet.bind"
    TAG = "bind"

    RECEIVE_NEW_FEATURES = False

    REQUIRED_PERMISSIONS: List[XmppPermission] = [
        XmppPermission.SEND_XML,
        XmppPermission.RECV_XML,
        XmppPermission.SET_JID,
        XmppPermission.SET_RESOURCE,
    ]

    def __init__(self, resource:str):
        self.__resource = resource
    
    def _connect_ci(self, ci):
        self.__ci = ci
    
    def _process(self, element):
        bind_xml = XmlElement(
            "iq",

            attributes = [
                XmlAttribute("type", "set"),
                XmlAttribute("id", "bind_1"),
            ],

            children = [
                XmlElement(
                    "bind",

                    attributes = [
                        XmlAttribute("xmlns", "urn:ietf:params:xml:ns:xmpp-bind")
                    ],

                    children = [
                        XmlElement(
                            "resource",
                            children = [
                                XmlTextElement(self.__resource)
                            ]
                        )
                    ]
                )
            ]
        )

        logger.debug(f"Sending bind request...")

        self.__ci.send_xml(bind_xml)
        data = self.__ci.recv_xml()

        if data.name != "iq":
            return None
        
        if data.get_attribute_by_name("type").value != "result":
            return None

        if data.get_child_by_name("bind") is None:
            return None

        jid = data.get_child_by_name("bind").get_child_by_name("jid").children[0].text

        logger.debug(f"Binded to '{jid}'!")

        self.__ci.set_jid(jid)
        self.__ci.set_resource(self.__resource)



import uuid

from osmxml import XmlParser
from osmxml import XmlElement
from osmxml import XmlAttribute
from osmxml import XmlTextElement


class SubscriptionXml:
    @staticmethod
    def check_for_subscription(jid: str) -> XmlElement:
        xml_str = f"""
        <iq from='{jid}' type='get' id='{SubscriptionXml.make_id()}'>
          <query xmlns='jabber:iq:roster'/>
        </iq>        
        """
        return XmlParser.parse_elements(xml_str)[0]

    @staticmethod
    def check_for_subscription_filter(xml: XmlElement) -> bool:
        if (xml.name != "iq"):
            return False

        query = xml.get_child_by_name("query")
        if (query is None):
            return False

        query_xmlns = query.get_attribute_by_name("xmlns")
        if ((query_xmlns is None) or (query_xmlns.value != "jabber:iq:roster")):
            return False

        return True 

    @staticmethod
    def send_presence(jid_to: str) -> XmlElement:
        xml_str = f"""
        <presence to='{jid_to}'>
          <show>chat</show>
          <status>Available for OMEMO</status>
        </presence>
        """
        return XmlParser.parse_elements(xml_str)[0]

    @staticmethod
    def send_subscribe(jid_to: str) -> XmlElement:
        xml_str = f"""
        <presence type='subscribe' to='{jid_to}'>
          <status>Subscription</status>
        </presence>
        """
        return XmlParser.parse_elements(xml_str)[0]

    @staticmethod
    def send_subscribe_filter(xml: XmlElement) -> bool:
        if (xml.name != "presence"):
            return False

        presence_type = xml.get_attribute_by_name("type")
        if ((presence_type is None) or (presence_type.value != "subscribe")):
            return False

        return True 

    @staticmethod
    def send_subscribed(jid_to: str) -> XmlElement:
        xml_str = f"""
        <presence type='subscribed' to='{jid_to}'/>
        """
        return XmlParser.parse_elements(xml_str)[0]

    @staticmethod
    def make_id():
        return str(uuid.uuid4()) 




from abc import ABC, abstractmethod


class XmppExtension(ABC):
    """
    Extensions are used to implement specific XMPP extensions & etc.
    They have more wider usecase than features.

    Attributes:
        ID (str): The ID of the extension implementation.
    """

    ID = None
    
    @abstractmethod
    def _connect_ci(self, ci) -> None:
        """
        Connects the extension to the client interface.

        Args:
            ci (XmppClientInterface): The client interface.
        """
        ...
    
    @abstractmethod
    def _process(self) -> None:
        """
        Processes the extension.
        """
        ...